local _G = _G
local Cartographer3 = _G.Cartographer3
local L = Cartographer3.L

function Cartographer3.MakeSuboptionsPanel(name, controlCreationFunc)
	local panel = CreateFrame("Frame", "Cartographer3_Panel_" .. name)
	
	panel.name = name
	panel.controls = {}
	if name ~= "Cartographer3" then
		panel.parent = "Cartographer3"
	end
	
	function panel:okay()
		for control in pairs(self.controls) do
			control.oldValue = control.value
			if control.okayFunc then
				control:okayFunc()
			end
		end
	end

	function panel:cancel()
		for control in pairs(self.controls) do
			control:SetValue(control.oldValue)
			if control.cancelFunc then
				control:cancelFunc()
			end
		end
	end

	function panel:default()
		for control in pairs(self.controls) do
			control:SetValue(control.default)
			if control.defaultFunc then
				control:defaultFunc()
			end
		end
	end
	
	InterfaceOptions_AddCategory(panel)
	
	panel:SetScript("OnShow", function(self)
		self:SetScript("OnShow", nil)
		controlCreationFunc(self)
		controlCreationFunc = nil
	end)
end

Cartographer3.MakeSuboptionsPanel("Cartographer3", function(self)
	local title = self:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetText("Cartographer3")
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetPoint("TOPLEFT", 16, -16)
	
	local subText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	subText:SetText(L["These options allow you to configure Cartographer3"]) -- TODO: better text
	subText:SetNonSpaceWrap(true)
	subText:SetJustifyH("LEFT")
	subText:SetJustifyV("TOP")
	subText:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subText:SetPoint("RIGHT", -32, 0)
	
	local versionText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	local format, height, website
	if Cartographer3.versionType == "Development" then
		format = "You are running %s from the source code repository. This is for |cffff7f7fdevelopers only|r. It is recommended in nearly all cases to instead use an alpha version from %s" -- no need for localization
		height = 40
		website = "http://wow.curseforge.com/projects/cartographer3/"
	elseif Cartographer3.versionType == "Alpha" then
		format = L["Version %s: This is for alpha testers wanting to try out new features and have possibly bad bugfixes |cffff7f7fbefore they are ready|r. It is recommended to instead use a beta or release version from %s"]
		height = 40
		website = "curse.com"
	elseif Cartographer3.versionType == "Beta" then
		format = L["Version %s: This is a beta version for those who want to test new features and bugfixes before they are potentially ready. It is recommended to use a release version from %s"]
		height = 40
		website = "curse.com"
	elseif Cartographer3.versionType == "Release" then
		format = L["Version %s. There may be more up-to-date versions on %s"]
		height = 20
		website = "curse.com"
	end
	versionText:SetText(format:format(Cartographer3.version, website))
	versionText:SetNonSpaceWrap(true)
	versionText:SetJustifyH("LEFT")
	versionText:SetJustifyV("TOP")
	versionText:SetPoint("TOPLEFT", subText, "BOTTOMLEFT", 0, -8)
	versionText:SetPoint("RIGHT", -32, 0)
	versionText:SetHeight(height)
	
	local function makeZoomToMinimapTextureSlider()
		local zoomToMinimapTextureSlider = CreateFrame("Slider", self:GetName() .. "_ZoomToMinimapTextureSlider", self, "InterfaceOptionsSliderTemplate")
		self.controls[zoomToMinimapTextureSlider] = true
		_G[zoomToMinimapTextureSlider:GetName() .. "Text"]:SetText(L["Texture Change Zoom"])
		zoomToMinimapTextureSlider.tooltipText = L["The amount to zoom in before the map changes from minimap-style textures to artwork.\n\nNote: Setting this too far may cause severe framerate issues."]
		_G[zoomToMinimapTextureSlider:GetName() .. "Text"]:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
		_G[zoomToMinimapTextureSlider:GetName() .. "Low"]:SetText(L["Far"])
		_G[zoomToMinimapTextureSlider:GetName() .. "Low"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		_G[zoomToMinimapTextureSlider:GetName() .. "High"]:SetText(L["Near"])
		_G[zoomToMinimapTextureSlider:GetName() .. "High"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		zoomToMinimapTextureSlider.default = Cartographer3.Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE
		zoomToMinimapTextureSlider:SetMinMaxValues(0, 120)
		zoomToMinimapTextureSlider.oldValue = Cartographer3.db.zoomToMinimapTexture
		zoomToMinimapTextureSlider:SetValue(zoomToMinimapTextureSlider.oldValue)
		local oldSetValue = zoomToMinimapTextureSlider.SetValue
		function zoomToMinimapTextureSlider:SetValue(value)
			oldSetValue(self, value)
			self.value = value
			Cartographer3.db.zoomToMinimapTexture = value
		end
		zoomToMinimapTextureSlider:SetScript("OnValueChanged", function(self)
			self.value = self:GetValue()
			self:SetValue(self:GetValue())
		end)
		return zoomToMinimapTextureSlider
	end
	
	local function makeStrataDropDown()
		local strataDropDown = CreateFrame("Frame", self:GetName() .. "_StrataDropDown", self, "UIDropDownMenuTemplate")
		self.controls[strataDropDown] = true
		local label = strataDropDown:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
		label:SetText(L["Strata"])
		strataDropDown.tooltipText = L["What level on the interface the map will show up on. 'Fullscreen' is the highest possible level and 'Background' is the lowest."]
		label:SetPoint("BOTTOMLEFT", strataDropDown, "TOPLEFT", 16, 3)
		local stratas = {
			"FULLSCREEN", L["Fullscreen"],
			"DIALOG", L["Dialog"],
			"HIGH", L["High"],
			"MEDIUM", L["Medium"],
			"LOW", L["Low"],
			"BACKGROUND", L["Background"]
		}
		UIDropDownMenu_Initialize(strataDropDown, function()
			for i = 1, #stratas, 2 do
				local value = stratas[i]
				local text = stratas[i+1]

				local info = UIDropDownMenu_CreateInfo()
				info.text = text
				info.value = value
				info.checked = Cartographer3.db.strata == value
				info.func = strataDropDown.SetValue
				info.arg1 = strataDropDown
				info.arg2 = value
				UIDropDownMenu_AddButton(info)
			end
		end)
		UIDropDownMenu_SetWidth(90, strataDropDown)
		UIDropDownMenu_SetSelectedValue(strataDropDown, Cartographer3.db.strata)
		strataDropDown.default = "FULLSCREEN"
		strataDropDown.oldValue = Cartographer3.db.strata
		function strataDropDown:SetValue(value)
			self.value = value
			Cartographer3.SetFrameStrata(value)
			UIDropDownMenu_SetSelectedValue(strataDropDown, value)
		end
		strataDropDown:EnableMouse(true)
		strataDropDown:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
			GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, 1)
		end)
		strataDropDown:SetScript("OnLeave", function(self)
			GameTooltip:Hide()
		end)
		return strataDropDown
	end
	
	local openReportIssueBox, openDonateBox
	do
		local url
		local function makeURLFrame()
			makeURLFrame = nil
			local function bumpFrameLevels(frame, amount)
				frame:SetFrameLevel(frame:GetFrameLevel()+amount)
				local children = newList(frame:GetChildren())
				for _,v in ipairs(children) do
					bumpFrameLevels(v, amount)
				end
				children = del(children)
			end
			-- some code borrowed from Prat here
			StaticPopupDialogs["CARTOGRAPHER3_SHOW_URL"] = {
				text = not IsMacClient() and L["Press Ctrl-C to copy, then Alt-Tab out of the game, open your favorite web browser, and paste the link into the address bar."] or L["Press Cmd-C to copy, then Cmd-Tab out of the game, open your favorite web browser, and paste the link into the address bar."],
				button2 = ACCEPT,
				hasEditBox = 1,
				hasWideEditBox = 1,
				showAlert = 1, -- HACK : it's the only way I found to make de StaticPopup have sufficient width to show WideEditBox :(

				OnShow = function()
					local editBox = _G[this:GetName() .. "WideEditBox"]
					editBox:SetText(url)
					editBox:SetFocus()
					editBox:HighlightText(0)
					editBox:SetScript("OnTextChanged", function() StaticPopup_EditBoxOnTextChanged() end)

					local button = _G[this:GetName() .. "Button2"]
					button:ClearAllPoints()
					button:SetWidth(200)
					button:SetPoint("CENTER", editBox, "CENTER", 0, -30)

					_G[this:GetName() .. "AlertIcon"]:Hide()  -- HACK : we hide the false AlertIcon
					this:SetFrameStrata("FULLSCREEN_DIALOG")
					this:SetFrameLevel(this:GetFrameLevel()+30)
				end,
				OnHide = function()
					local editBox = _G[this:GetName() .. "WideEditBox"]
					editBox:SetScript("OnTextChanged", nil)
					this:SetFrameStrata("DIALOG")
					this:SetFrameLevel(this:GetFrameLevel()-30)
				end,
				OnAccept = function() end,
				OnCancel = function() end,
				EditBoxOnEscapePressed = function() this:GetParent():Hide() end,
				EditBoxOnTextChanged = function()
					this:SetText(url)
					this:SetFocus()
					this:HighlightText(0)
				end,
				timeout = 0,
				whileDead = 1,
				hideOnEscape = 1
			}
		end
		
		function openReportIssueBox()
			if makeURLFrame then
				makeURLFrame()
			end
		
			url = "http://wow.curseforge.com/projects/cartographer3/tickets/"

			StaticPopup_Show("CARTOGRAPHER3_SHOW_URL")
		end
		
		function openDonateBox()
			if makeURLFrame then
				makeURLFrame()
			end
			
			url = "https://www.paypal.com/cgi-bin/webscr?cmd=_xclick&business=ckknight%40gmail%2ecom&item_name=Cartographer3"

			StaticPopup_Show("CARTOGRAPHER3_SHOW_URL")
		end
	end
	
	local function makeReportButton()
		local reportButton = CreateFrame("Button", self:GetName() .. "_ReportButton", self, "UIPanelButtonTemplate2")
		self.controls[reportButton] = true
		reportButton:SetText(L["Report Issue"])
		reportButton.tooltipText = L["Report a defect or a possible enhancement you see with Cartographer3."]
		reportButton:SetWidth(120)
		reportButton:SetHeight(22)
		function reportButton:SetValue()
		end
		reportButton:SetScript("OnClick", function()
			openReportIssueBox()
		end)
		reportButton:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, 1)
		end)
		reportButton:SetScript("OnLeave", function(self)
			GameTooltip:Hide()
		end)
		return reportButton
	end
	
	local function makeDonateButton()
		local donateButton = CreateFrame("Button", self:GetName() .. "_DonateButton", self, "UIPanelButtonTemplate2")
		self.controls[donateButton] = true
		donateButton:SetText(L["Donate"])
		donateButton.tooltipText = L["Give a much-needed donation to ckknight, the author of Cartographer3"]
		donateButton:SetWidth(120)
		donateButton:SetHeight(22)
		function donateButton:SetValue()
		end
		donateButton:SetScript("OnClick", function()
			openDonateBox()
		end)
		donateButton:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, 1)
		end)
		donateButton:SetScript("OnLeave", function(self)
			GameTooltip:Hide()
		end)
		return donateButton
	end
	
	local function makeCloseWithEscapeToggle()
		local closeWithEscapeToggle = CreateFrame("CheckButton", self:GetName() .. "_CloseWithEscapeToggle", self, "InterfaceOptionsCheckButtonTemplate")
		self.controls[closeWithEscapeToggle] = true
		_G[closeWithEscapeToggle:GetName() .. "Text"]:SetText(L["Close map with escape key"])
		closeWithEscapeToggle.default = true
		closeWithEscapeToggle.oldValue = Cartographer3.db.closeWithEscape
		function closeWithEscapeToggle:SetValue(value)
			Cartographer3.db.closeWithEscape = not not value
			self.value = value
			self:SetChecked(value)
		end
		closeWithEscapeToggle:SetScript("OnClick", function(self)
			self:SetValue(not not self:GetChecked())
		end)
		closeWithEscapeToggle:SetChecked(closeWithEscapeToggle.oldValue)
		closeWithEscapeToggle:SetScript("OnEnter", function(this)
			GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
			GameTooltip:SetText(L["Set whether pressing the escape key will close the map or not."], nil, nil, nil, nil, 1)
		end)
		closeWithEscapeToggle:SetScript("OnLeave", function(this)
			GameTooltip:Hide()
		end)
		return closeWithEscapeToggle
	end
	
	local function makeOpacitySlider()
		local opacitySlider = CreateFrame("Slider", self:GetName() .. "_OpacitySlider", self, "InterfaceOptionsSliderTemplate")
		self.controls[opacitySlider] = true
		_G[opacitySlider:GetName() .. "Text"]:SetText(L["Opacity"])
		opacitySlider.tooltipText = L["The opacity of the Cartographer3 map"]
		_G[opacitySlider:GetName() .. "Text"]:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
		_G[opacitySlider:GetName() .. "Low"]:SetText("10%")
		_G[opacitySlider:GetName() .. "Low"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		_G[opacitySlider:GetName() .. "High"]:SetText("100%")
		_G[opacitySlider:GetName() .. "High"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		opacitySlider.default = 1
		opacitySlider:SetMinMaxValues(0.1, 1)
		opacitySlider.oldValue = Cartographer3.db.opacity
		opacitySlider:SetValue(opacitySlider.oldValue)
		local oldSetValue = opacitySlider.SetValue
		function opacitySlider:SetValue(value)
			oldSetValue(self, value)
			self.value = value
			Cartographer3.SetOpacity(value)
		end
		opacitySlider:SetScript("OnValueChanged", function(self)
			self.value = self:GetValue()
			self:SetValue(self:GetValue())
		end)
		return opacitySlider
	end
	
	local function makeUnexploredColorPicker()
		local button = CreateFrame("Button", self:GetName() .. "_UnexploredColorPicker", self)
		self.controls[button] = true
		button:SetHitRectInsets(0, -100, 0, 0)
		local text = button:CreateFontString(nil, "ARTWORK", "GameFontHighlight")
		text:SetText(L["Unexplored color"])
		text:SetPoint("LEFT", button, "RIGHT", 0, 1)
		local color = button:CreateTexture(nil, "ARTWORK")
		button.color = color
		color:SetTexture(unpack(Cartographer3.db.unexploredColor))
		local background = button:CreateTexture(nil, "BORDER")
		background:SetTexture([=[Tileset\Generic\Checkers]=])
		background:SetTexCoord(0, 0.5, 0, 0.5)
		local border = button:CreateTexture(nil, "BACKGROUND")
		border:SetTexture([=[Interface\ChatFrame\ChatFrameColorSwatch]=])
		button:SetWidth(26)
		button:SetHeight(26)
		background:SetPoint("CENTER")
		background:SetWidth(14)
		background:SetHeight(14)
		color:SetPoint("CENTER")
		color:SetWidth(14)
		color:SetHeight(14)
		border:SetAllPoints(button)
		button.default = { 1, 1, 1, 1 }
		button.oldValue = { unpack(Cartographer3.db.unexploredColor) }
		local info
		local function func(r, g, b, a)
			info.r = r
			info.g = g
			info.b = b
			info.opacity = a
			color:SetTexture(r, g, b, a)
			if not button.value or button.value == button.oldValue then
				button.value = { r, g, b, a }
			else
				button.value[1] = r
				button.value[2] = g
				button.value[3] = b
				button.value[4] = a
			end
			info.r = r
			info.g = g
			info.b = b
			info.opacity = 1 - a
			Cartographer3.SetUnexploredColor(r, g, b, a)
		end
		function button:SetValue(value)
			func(unpack(value))
		end
		local function swatchFunc()
			local r, g, b = ColorPickerFrame:GetColorRGB()
			local opacity = 1 - OpacitySliderFrame:GetValue()
			
			func(r, g, b, opacity)
		end
		info = {
			swatchFunc = swatchFunc,
			hasOpacity = true,
			opacityFunc = swatchFunc,
			r = Cartographer3.db.unexploredColor[1],
			g = Cartographer3.db.unexploredColor[2],
			b = Cartographer3.db.unexploredColor[3],
			opacity = 1 - Cartographer3.db.unexploredColor[4],
			cancelFunc = function(...)
				local previousValues = ColorPickerFrame.previousValues
				local r, g, b, opacity = previousValues.r, previousValues.g, previousValues.b, 1 - previousValues.opacity
				
				func(r, g, b, opacity)
			end,
		}
		button:SetScript("OnClick", function(this)
			OpenColorPicker(info)
		end)
		button:SetScript("OnEnter", function(this)
			GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
			GameTooltip:SetText(L["Set the color of unexplored areas on the Cartographer3 map. White shows everything and black obscures everything."], nil, nil, nil, nil, 1)
		end)
		button:SetScript("OnLeave", function(this)
			GameTooltip:Hide()
		end)
		button:RegisterForClicks("LeftButtonUp")
		return button
	end
	
	local reportButton = makeReportButton()
	reportButton:SetPoint("TOPLEFT", versionText, "BOTTOMLEFT", 0, -16)
	
	local donateButton = makeDonateButton()
	donateButton:SetPoint("TOPRIGHT", versionText, "BOTTOMRIGHT", 0, -16)
	
	local zoomToMinimapTextureSlider = makeZoomToMinimapTextureSlider()
	zoomToMinimapTextureSlider:SetPoint("TOPLEFT", reportButton, "BOTTOMLEFT", 0, -32)
	
	local strataDropDown = makeStrataDropDown()
	strataDropDown:SetPoint("TOPRIGHT", donateButton, "BOTTOMRIGHT", 13, -28)
	
	local closeWithEscapeToggle = makeCloseWithEscapeToggle()
	closeWithEscapeToggle:SetPoint("TOPLEFT", zoomToMinimapTextureSlider, "BOTTOMLEFT", 0, -24)
	
	local opacitySlider = makeOpacitySlider()
	opacitySlider:SetPoint("TOPLEFT", closeWithEscapeToggle, "BOTTOMLEFT", 0, -24)
	
	local unexploredColorPicker = makeUnexploredColorPicker()
	unexploredColorPicker:SetPoint("TOPLEFT", opacitySlider, "BOTTOMLEFT", 0, -32)
end)

Cartographer3.MakeSuboptionsPanel(L["POIs"], function(self)
	local title = self:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetText(L["POIs"])
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetPoint("TOPLEFT", 16, -16)

	local subText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	subText:SetText(L["These are options to change how POIs look"])
	subText:SetNonSpaceWrap(true)
	subText:SetJustifyH("LEFT")
	subText:SetJustifyV("TOP")
	subText:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subText:SetPoint("RIGHT", -32, 0)
	
	local last = subText
	
	local t = {}
	for id in pairs(Cartographer3.Data.poiTypes) do
		t[#t+1] = id
	end
	table.sort(t)
	for _, id in ipairs(t) do
		local localizedName = Cartographer3.Data.poiTypes[id]
		local exampleFrame = Cartographer3.Data.poiTypeExampleFrames[id]()
		Cartographer3.Data.poiTypeExampleFrames[id] = nil
	
		local sizeSlider = CreateFrame("Slider", self:GetName() .. "_SizeSlider_" .. id, self, "InterfaceOptionsSliderTemplate")
		self.controls[sizeSlider] = true
		_G[sizeSlider:GetName() .. "Text"]:SetText(localizedName)
		sizeSlider.tooltipText = L["The size that %s appears on the map"]:format(localizedName)
		_G[sizeSlider:GetName() .. "Text"]:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
		_G[sizeSlider:GetName() .. "Low"]:SetText(L["Small"])
		_G[sizeSlider:GetName() .. "Low"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		_G[sizeSlider:GetName() .. "High"]:SetText(L["Large"])
		_G[sizeSlider:GetName() .. "High"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		sizeSlider.default = 1
		sizeSlider:SetMinMaxValues(0.25, 4)
		sizeSlider.oldValue = Cartographer3.db.pois[id]
		sizeSlider:SetValue(sizeSlider.oldValue)
		local oldSetValue = sizeSlider.SetValue
		
		local function resize(value)
			local scale = Cartographer3.Data.poiTypeScales[id]
			exampleFrame:SetWidth(scale * 20 * value)
			exampleFrame:SetHeight(scale * 20 * value)
		end
		
		function sizeSlider:SetValue(value)
			oldSetValue(self, value)
			self.value = value
			Cartographer3.db.pois[id] = value
			for poi in pairs(Cartographer3.Data.shownPOIs) do
				poi:Resize()
			end
			resize(value)
		end
		sizeSlider:SetScript("OnValueChanged", function(self)
			self.value = self:GetValue()
			self:SetValue(self:GetValue())
		end)
		
		sizeSlider:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -24)
		
		exampleFrame:SetParent(self)
		exampleFrame:SetPoint("CENTER", sizeSlider, "RIGHT", 100, 0)
		
		resize(sizeSlider.oldValue)
		
		last = sizeSlider
	end
end)

Cartographer3.MakeSuboptionsPanel(L["Advanced"], function(self)
	local title = self:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetText(L["Advanced"])
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetPoint("TOPLEFT", 16, -16)
	
	local subText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	subText:SetText(L["These are advanced options for Cartographer3"])
	subText:SetNonSpaceWrap(true)
	subText:SetJustifyH("LEFT")
	subText:SetJustifyV("TOP")
	subText:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subText:SetPoint("RIGHT", -32, 0)
	
	local function makeHijackWorldMapToggle()
		local hijackWorldMapToggle = CreateFrame("CheckButton", self:GetName() .. "_HijackWorldMapToggle", self, "InterfaceOptionsCheckButtonTemplate")
		self.controls[hijackWorldMapToggle] = true
		_G[hijackWorldMapToggle:GetName() .. "Text"]:SetText(L["Separate Cartographer3 from the World Map"])
		hijackWorldMapToggle.default = false
		hijackWorldMapToggle.oldValue = not Cartographer3.db.hijackWorldMap
		function hijackWorldMapToggle:SetValue(value)
			Cartographer3.db.hijackWorldMap = not value
			self.value = value
			self:SetChecked(value)
		end
		hijackWorldMapToggle:SetScript("OnClick", function(self)
			self:SetValue(not not self:GetChecked())
		end)
		hijackWorldMapToggle:SetChecked(hijackWorldMapToggle.oldValue)
		hijackWorldMapToggle:SetScript("OnEnter", function(this)
			GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
			GameTooltip:SetText(L["Set whether to separate Cartographer3 from the World Map.\n\nNote: This will prevent most third-party addons from working properly.\n\nYou can set the keybinding to open and close Cartographer3 in the Keybindings menu.\n\nChanging this initiates a |cffff0000ReloadUI|r"], nil, nil, nil, nil, 1)
		end)
		hijackWorldMapToggle:SetScript("OnLeave", function(this)
			GameTooltip:Hide()
		end)
		function hijackWorldMapToggle:okayFunc()
			if Cartographer3.db.hijackWorldMap ~= Cartographer3.hijackingWorldMap then
				ReloadUI()
			end
		end
		hijackWorldMapToggle.cancelFunc = hijackWorldMapToggle.okayFunc
		hijackWorldMapToggle.defaultFunc = hijackWorldMapToggle.okayFunc
		return hijackWorldMapToggle
	end
	
	local hijackWorldMapToggle = makeHijackWorldMapToggle()
	hijackWorldMapToggle:SetPoint("TOPLEFT", subText, "BOTTOMLEFT", 0, -16)
end)
