local Cartographer3 = _G.Cartographer3
if not Cartographer3 then
	error("Cartographer3_Waypoints requires Cartographer3")
end

local L = Cartographer3.L

local Cartographer3_Waypoints = {}
Cartographer3.Waypoints = Cartographer3_Waypoints
_G.Cartographer3_Waypoints = Cartographer3_Waypoints -- add a global for easy checking from 3rd parties

local Cartographer3_Utils = Cartographer3.Utils

local db

local updateArrow

local waypoint_x, waypoint_y, waypoint_text, waypoint_metadata

function Cartographer3_Waypoints.SetWaypoint(zone, x, y, text, metadata)
	local ux, uy = Cartographer3_Utils.ConvertZoneCoordinateToUniverseCoordinate(zone, x, y)
	
	waypoint_x = ux
	waypoint_y = uy
	waypoint_text = text
	waypoint_metadata = metadata
	
	updateArrow(0, GetTime())
end
_G.SetWaypoint = Cartographer3_Waypoints.SetWaypoint

function Cartographer3_Waypoints.GetWaypoint()
	return waypoint_metadata
end
_G.GetWaypoint = Cartographer3_Waypoints.GetWaypoint

function Cartographer3_Waypoints.ClearWaypoint()
	waypoint_x = nil
	waypoint_y = nil
	waypoint_text = nil
	waypoint_metadata = nil
	
	updateArrow(0, GetTime())
end
_G.ClearWaypoint = Cartographer3_Waypoints.ClearWaypoint

do
	local zoneTexture, zoneX, zoneY
	local function setWaypointHere()
		Cartographer3_Waypoints.SetWaypoint(zoneTexture, zoneX, zoneY)
	end
	Cartographer3.AddRightClickMenuHandler(function(data, level, value)
		if level == 1 then
			local info = UIDropDownMenu_CreateInfo()
			info.text = L["Set waypoint here"]
			info.func = setWaypointHere
			zoneTexture, zoneX, zoneY = data.zoneTexture, data.zoneX, data.zoneY
			UIDropDownMenu_AddButton(info, level)
		end
	end)
end

do
	local button = CreateFrame("Button", "Cartographer3_Waypoints_Button", UIParent)
	button:SetWidth(100)
	button:SetHeight(55)
	button:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	button:EnableMouse(true)
	button:SetMovable(true)
	button:RegisterForDrag("LeftButton")
	button:SetScript("OnDragStart", function(this)
		this:StartMoving()
	end)
	button:SetScript("OnDragStop", function(this)
		this:StopMovingOrSizing()
	end)
	button:RegisterForClicks("RightButtonUp")
	button:Hide()
	button:SetClampedToScreen(true)
	local dropdownFrame
	local menuList
	local locked = false
	button:SetScript("OnClick", function(this)
		if not dropdownFrame then
			dropdownFrame = CreateFrame("Frame", this:GetName() .. "_Dropdown", this, "UIDropDownMenuTemplate")
			-- Clear waypoint
			-- Lock arrow
			menuList = {
				{
					text = L["Cartographer3 Waypoint Arrow"],
					isTitle = true,
				},
				{
					text = L["Clear waypoint"],
					func = Cartographer3_Waypoints.ClearWaypoint,
				},
				{
					text = L["Lock arrow position"],
					func = function()
						locked = not locked
						if locked then
							button:SetMovable(false)
							button:RegisterForDrag()
						else
							button:SetMovable(true)
							button:RegisterForDrag("LeftButton")
						end
						menuList[3].checked = locked
					end,
					checked = locked
				},
				{
					text = " ",
					isTitle = true,
				},
				{
					text = _G.CLOSE,
					func = function() end,
				}
			}
		end
		EasyMenu(menuList, dropdownFrame, "cursor", 0, 0, "MENU")
	end)
	
	local arrow = CreateFrame("PlayerModel", "Cartographer3_Waypoints_ArrowModel", button)
	arrow:SetModel([=[Interface\AddOns\Cartographer3\WaypointArrow.m2]=])
	arrow:SetModelScale(1.55)
	arrow:SetLight(1, 0, 0, 1, 0, 1, 0.7, 0.7, 0.7, 1, 0.8, 0.8, 0.64)
	arrow:SetPoint("CENTER")
	arrow:SetWidth(button:GetWidth())
	arrow:SetHeight(button:GetHeight())
	
	local text = arrow:CreateFontString(nil, "ARTWORK", "GameFontHighlight")
	arrow.text = text
	text:SetPoint("TOP", arrow, "BOTTOM")
	
	local yardage = arrow:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	arrow.yardage = yardage
	yardage:SetPoint("TOP", text, "BOTTOM")
	
	local time = arrow:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	arrow.time = time
	time:SetPoint("TOP", yardage, "BOTTOM")
	
	local YARDS_PER_PIXEL = Cartographer3.Data.YARDS_PER_PIXEL
	
	local last_playerX, last_playerY
	local lastUpdate = 0
	function updateArrow(elapsed, currentTime)
		if not waypoint_x then
			button:Hide()
			return
		end
		
		local playerX, playerY = Cartographer3.Utils.GetUnitUniverseCoordinate("player")
		if not playerX then
			button:Hide()
			return
		end
		
		local destinationDeltaX, destinationDeltaY = waypoint_x - playerX, waypoint_y - playerY
		
		local angleToDestination = math.atan2(destinationDeltaY, destinationDeltaX)
		
		local playerRotation = Cartographer3.Utils.GetPlayerRotation()
		
		local arrowRotation = angleToDestination + playerRotation
		arrow:SetRotation(arrowRotation)
		
		local value = (((arrowRotation / math.pi) % 2) - 1)^2
		
		if value >= 0.5 then
			r, g, b = 2 - value*2, 1, 0
		else
			r, g, b = 1, value*2, 0
		end
		arrow:SetLight(1, 0, 0, 0.2, -0.8, 0.5, r, g, b, 0.8, r, g, b)
		
		arrow.text:SetText(waypoint_text)
		
		local yardsToDestination = (destinationDeltaX*destinationDeltaX + destinationDeltaY*destinationDeltaY)^0.5 * YARDS_PER_PIXEL
		
		if yardsToDestination < 5 then -- allow changing of yards
			Cartographer3_Waypoints.ClearWaypoint()
			return
		end
		
		local yardString = ("%.0f yd"):format(yardsToDestination)
		arrow.yardage:SetText(yardString)
		
		local diffTime = currentTime - lastUpdate
		if diffTime >= 1 then
			lastUpdate = currentTime
			if not last_playerX then
				last_playerX, last_playerY = playerX, playerY
				return
			end
			
			local movementX, movementY = playerX - last_playerX, playerY - last_playerY
			local movementAngle = math.atan2(movementY, movementX)
			local movementYardsPerSecond = (movementX*movementX + movementY*movementY)^0.5 * YARDS_PER_PIXEL / diffTime
			movementYardsPerSecond = math.floor(movementYardsPerSecond * 100 + 0.5) / 100
			
			local secondsToDestination = yardsToDestination / movementYardsPerSecond
			
			local adjustedSecondsToDestination = secondsToDestination / math.cos(movementAngle - angleToDestination)
			
			local timeString
			if adjustedSecondsToDestination < 0 or adjustedSecondsToDestination > 24*60*60 then
				timeString = '---'
			else
				if adjustedSecondsToDestination > 60*60 then
					timeString = ("%d:%02d:%02d"):format(adjustedSecondsToDestination / (60*60), (adjustedSecondsToDestination % (60*60)) / 60, adjustedSecondsToDestination % 60)
				else
					timeString = ("%d:%02d"):format(adjustedSecondsToDestination / 60, adjustedSecondsToDestination % 60)
				end
			end
		
			arrow.time:SetText(timeString)
			last_playerX, last_playerY = playerX, playerY
		end
		
		if not button:IsShown() then
			arrow:SetModel([=[Interface\AddOns\Cartographer3\WaypointArrow.m2]=])
			button:Show()
		end	
	end
	Cartographer3_Utils.AddTimer(updateArrow, true)
end

function Cartographer3_Waypoints.SetArrowSize(value)
	db.arrowSize = value
	Cartographer3_Waypoints_Button:SetScale(value)
	Cartographer3_Waypoints_ArrowModel:SetModelScale(1.55 / value)
end

function Cartographer3_Waypoints.OnInitialize()
	if not Cartographer3.db.Waypoints then
		Cartographer3.db.Waypoints = {}
	end
	db = Cartographer3.db.Waypoints
	if not db.arrowSize then
		db.arrowSize = 1
	end
	Cartographer3_Waypoints.SetArrowSize(db.arrowSize)
end

Cartographer3.MakeSuboptionsPanel(L["Waypoints"], function(self)
	local title = self:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetText(("Cartographer3 - %s"):format(L["Waypoints"]))
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetPoint("TOPLEFT", 16, -16)

	local subText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	subText:SetText(L["These options allow you to configure Cartographer3's waypoints"]) -- TODO: better text
	subText:SetNonSpaceWrap(true)
	subText:SetJustifyH("LEFT")
	subText:SetJustifyV("TOP")
	subText:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subText:SetPoint("RIGHT", -32, 0)
	
	local function makeArrowSizeSlider()
		local arrowSizeSlider = CreateFrame("Slider", self:GetName() .. "_ArrowSizeSlider", self, "InterfaceOptionsSliderTemplate")
		self.controls[arrowSizeSlider] = true
		_G[arrowSizeSlider:GetName() .. "Text"]:SetText(L["Arrow Size"])
		arrowSizeSlider.tooltipText = L["The size of the waypoint arrow"]
		_G[arrowSizeSlider:GetName() .. "Text"]:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
		_G[arrowSizeSlider:GetName() .. "Low"]:SetText("25%")
		_G[arrowSizeSlider:GetName() .. "Low"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		_G[arrowSizeSlider:GetName() .. "High"]:SetText("400%")
		_G[arrowSizeSlider:GetName() .. "High"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
		arrowSizeSlider.default = Cartographer3.Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE
		arrowSizeSlider:SetMinMaxValues(0.25, 4)
		arrowSizeSlider.oldValue = db.arrowSize
		arrowSizeSlider:SetValue(arrowSizeSlider.oldValue)
		local oldSetValue = arrowSizeSlider.SetValue
		function arrowSizeSlider:SetValue(value)
			oldSetValue(self, value)
			self.value = value
			Cartographer3_Waypoints.SetArrowSize(value)
		end
		arrowSizeSlider:SetScript("OnValueChanged", function(self)
			self.value = self:GetValue()
			self:SetValue(self:GetValue())
		end)
		return arrowSizeSlider
	end
	
	local arrowSizeSlider = makeArrowSizeSlider()
	arrowSizeSlider:SetPoint("TOPLEFT", subText, "BOTTOMLEFT", 0, -32)
end)


for i, v in ipairs { "/waypoint", "/way", "/tway", "/tloc" } do
	_G["SLASH_CARTOGRAPHERTHREE_WAYPOINTS" .. i] = v
end

local function printUsage()
	DEFAULT_CHAT_FRAME:AddMessage(L["|cffffff7fCartographer3 - Waypoints |r/way |cffffff7fUsage:|r"])
	DEFAULT_CHAT_FRAME:AddMessage(L["|cffffff7f/way <x> <y> [description]|r - Set waypoint to x, y with description"])
	DEFAULT_CHAT_FRAME:AddMessage(L["|cffffff7f/way <zone> <x> <y> [description]|r - Set waypoint to x, y in zone with description"])
	DEFAULT_CHAT_FRAME:AddMessage(L["|cffffff7f/way clear|r - Clear waypoint"])
end

local zoneList
local possibleMatches = {}
_G.hash_SlashCmdList.CARTOGRAPHERTHREE_WAYPOINTS = nil
_G.SlashCmdList.CARTOGRAPHERTHREE_WAYPOINTS = function(text)
	local tokens = { (", "):split(text) }
	
	for i = #tokens, 1, -1 do
		local v = tokens[i]
		if v:trim() == "" then
			table.remove(tokens, i)
		end
	end
	
	local first = tokens[1]:lower()
	
	if first == "clear" then
		Cartographer3_Waypoints.ClearWaypoint()
	elseif first ~= "" then
		local num = 0
		for i, v in ipairs(tokens) do
			if tonumber(v) then
				num = i - 1
				break
			end
		end
		
		local zone = nil
		if num > 0 then
			zone = table.concat(tokens, " ", 1, num)
		end
		
		local x, y = tokens[num + 1], tokens[num + 2]
		
		local name = table.concat(tokens, " ", num + 3)
		if name == "" then
			name = nil
		end
		
		local texture
		if zone then
			if not zoneList then
				zoneList = {}
				for texture, zoneData in pairs(Cartographer3.Data.ZONE_DATA) do
					zoneList[zoneData.localizedName:lower():gsub("[%L]", "")] = texture
				end
			end
			local lowerZone = zone:lower():gsub("[%L]", "")
			for name, texture in pairs(zoneList) do
				if name:match(lowerZone) then
					possibleMatches[#possibleMatches+1] = texture
				end
			end
			if #possibleMatches > 1 then
				local names = {}
				for i, texture in ipairs(possibleMatches) do
					names[#names+1] = Cartographer3.Data.ZONE_DATA[texture].localizedName
				end
				table.sort(names)
				
				DEFAULT_CHAT_FRAME:AddMessage(L["Found multiple matches for zone '%s': %s"]:format(zone, table.concat(names, ", ")))
				for i = 1, #possibleMatches do
					possibleMatches[i] = nil
				end
				return
			end
			texture = possibleMatches[1]
			possibleMatches[1] = nil
			
			if not texture then
				DEFAULT_CHAT_FRAME:AddMessage(L["No match was found for zone '%s'"]:format(zone))
				return
			end
		end
		
		x, y = tonumber(x), tonumber(y)
		if not x or not y then
			printUsage()
			return
		end
		
		x, y = x / 100, y / 100
		
		if not texture then
			texture = GetMapInfo()
		end
		
		Cartographer3_Waypoints.SetWaypoint(texture, x, y, name)
	else
		printUsage()
	end
end
